
#include <fcntl.h>
#include <niche/nadriver.h>
#include <signal.h>
#include <sys/time.h>
#include "qserver.h"

#define MAX_LINK_TIME           2000   /* millisecs */

extern char main_buffer [BUFFER_SIZE] ;

int link_descriptor ;
int link_mask ;

int boot_network (boot_file, device_name)
char *boot_file, *device_name ;
{
  link_descriptor = open (device_name, O_RDWR) ;
  if (link_descriptor >= 0)
  {
    int file_descriptor ;

    link_mask = 1 << link_descriptor ;
    ioctl (link_descriptor, SITE_RESET) ;
    ioctl (link_descriptor, DISABLE_ERRORS) ;
    file_descriptor = open (boot_file, O_RDONLY) ;
    if (file_descriptor >= 0)
    {
      int eof = 0 ;
      while (! eof)
      {
        int bytes_read = read (file_descriptor, main_buffer, BUFFER_SIZE) ;
        if (bytes_read > 0)
        {
          if (! output_or_fail (link_descriptor, main_buffer, bytes_read))
          {
            printf ("Failed to boot network\n") ;
            return (0) ;
          }
        }
        else
          if (bytes_read == 0)
            eof = 1 ;
          else
          {
            printf ("Error reading boot file\n") ;
            return (0) ;
          }
      }
      ioctl (link_descriptor, ENABLE_ERRORS) ;
      printf ("Network booted ok\n") ;
      return (1) ;
    }
    else 
    {
      printf ("Failed to open boot-file \"%s\"\n", boot_file) ;
      return (0) ;
    }
  }
  else
  {
    printf ("Device \"%s\" is not available\n", device_name) ;
    return (0) ;
  }
}

int output_or_fail (device_descriptor, buffer, len)
int device_descriptor ;
char *buffer ;
int len ;
/*
-- This seems rather inefficient, but I don't know any other
-- way to do it.

-- Spawn two child(ren) processes.  One attempts the write then exits,
-- the other pauses for some interval of time (the time-out interval)
-- then exits.
-- Depending on which process terminates first the output
-- has succeeded or failed.

-- Note: As coded, for this to succeed, the calling process must
-- not have any already existing children at the time of the call.
*/
{
  int buffer_id, guard_id, result ;

  if (buffer_id = fork ())
  {
    if (guard_id = fork ())
    {
      /* The parent process */

      int winner ;

      winner = wait (0) ;
      if (winner == buffer_id)
      {
        kill (guard_id, SIGKILL) ;
        result = 1 ;
      }
      else
        if (winner == guard_id)
        {
          kill (buffer_id, SIGKILL) ;
          printf ("Time-out on output\n") ;
          result = 0 ;
        }
        else
        {
          kill (buffer_id, SIGKILL) ;
          kill (guard_id,  SIGKILL) ;
          printf ("\'wait\' error.  Result = %d\n", winner) ;
          result = 0 ;
        }
    }
    else
    {
      /* The time-out guard process */
 
      struct timeval timelimit ;
      int result ;

      timelimit.tv_sec  = MAX_LINK_TIME / 1000 ;
      timelimit.tv_usec = 1000 * (MAX_LINK_TIME % 1000);
      result = select (32, 0, 0, 0, &timelimit) ;
      exit (0) ;
    }
  }
  else
  {
    /* The output buffer process */

    int result ;

    result = write (device_descriptor, buffer, len) ;
    exit (0) ;
  }

  /*
  Make sure all children have really stopped, ready for
  next call.
  */
  while (wait (0) >= 0)
    ;

  return (result) ;
}

void link_receive (data, len)
char *data ;
int len ;
{
  int result ;
  int residue  = len  ;

  while (residue)
    if (residue <= LINK_IN_PAK_SIZE)
    {
      result   = read (link_descriptor, data, residue) ;
      residue  = 0 ;
    }
    else
    {
      result   = read (link_descriptor, data, LINK_IN_PAK_SIZE) ;
      residue -= LINK_IN_PAK_SIZE ;
      data    += LINK_IN_PAK_SIZE ;
    }
}

void link_send (data, len)
char *data ;
int len ;
{
  int result ;
  int residue = len ;

  while (residue)
    if (residue <= LINK_OUT_PAK_SIZE)
    {
      result   = write (link_descriptor, data, residue) ;
      residue  = 0 ;
    }
    else
    {
      result   = write (link_descriptor, data, LINK_OUT_PAK_SIZE) ;
      residue -= LINK_OUT_PAK_SIZE ;
      data    += LINK_OUT_PAK_SIZE ;
    }
}

